/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#include "DirichletBC.h"
#include "ElInfo.h"
#include "BasisFunction.h"
#include "DOFVector.h"
#include "DOFMatrix.h"

namespace AMDiS {

  DirichletBC::DirichletBC(BoundaryType type,
			   AbstractFunction<double, WorldVector<double> > *fct,
			   const FiniteElemSpace *rowFeSpace,
			   const FiniteElemSpace *colFeSpace,
			   bool apply)
    : BoundaryCondition(type, rowFeSpace, colFeSpace), 
      f(fct), 
      dofVec(NULL),
      applyBC(apply)
  {}


  DirichletBC::DirichletBC(BoundaryType type,
			   DOFVectorBase<double> *vec,
			   bool apply)
    : BoundaryCondition(type, vec->getFeSpace(), vec->getFeSpace()), 
      f(NULL), 
      dofVec(vec),
      applyBC(apply)
  {}


  void DirichletBC::fillBoundaryCondition(DOFMatrix* matrix,
					  ElInfo* elInfo,
					  const DegreeOfFreedom* dofIndices,
					  const BoundaryType* localBound,
					  int nBasFcts)
  {
    FUNCNAME_DBG("DirichletBC::fillBoundaryCondition()");
    TEST_EXIT_DBG(matrix->getRowFeSpace() == rowFeSpace)("invalid row fe space\n");
  }


  void DirichletBC::fillBoundaryCondition(DOFVectorBase<double>* vector,
					  ElInfo* elInfo,
					  const DegreeOfFreedom* dofIndices,
					  const BoundaryType* localBound,
					  int nBasFcts)
  {
    FUNCNAME("DirichletBC::fillBoundaryCondition()");

    TEST_EXIT_DBG(vector->getFeSpace() == rowFeSpace)("invalid row fe space\n");

    const BasisFunction *basFcts = rowFeSpace->getBasisFcts();

    for (int i = 0; i < nBasFcts; i++) {
      if (localBound[i] == boundaryType) {
	double value = 0.0;
	if (f) {
	  elInfo->coordToWorld(*(basFcts->getCoords(i)), worldCoords);
	  value = (*f)(worldCoords);
	} else {
	  if (dofVec)
	    value = (*dofVec)[dofIndices[i]];
	  else
	    ERROR_EXIT("There is something wrong!\n");
	}
	vector->setDirichletDofValue(dofIndices[i], value);
	(*vector)[dofIndices[i]] = value;
      }
    }
  }


  void DirichletBC::initVector(DOFVectorBase<double>* vec)
  {
    if (dynamic_cast<DOFVector<double>*>(vec))
      dynamic_cast<DOFVector<double>*>(vec)->getDirichletValues().clear();
  }

}
