/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#ifndef EXTENSIONS_GRADIENT_CALCULATIONS_H
#define EXTENSIONS_GRADIENT_CALCULATIONS_H

#include "AMDiS.h"

namespace AMDiS { namespace extensions {

/// Extended recovery gradient assuming mirror/point symmetry (pointSym=false/true)
/// of the solution at domain boundaries

/// --> Simon, es hat sich herausgestellt, dass meine ursprüngliche Version mit *nrm
/// effizienter ist, da Speicher nur für einige Ausnahmeknoten belegt wird (ok, ist
/// vielleicht etwas unübersichtlich)

  inline void getRecoveryGradientSym(const DOFVector<double> *u,
				     DOFVector<WorldVector<double> > *result, 
				     bool pointSym = false)
  {
    FUNCNAME("getRecoveryGradientSym()");
    const FiniteElemSpace *feSpace = u->getFeSpace();
    int dim = feSpace->getMesh()->getDim();

    if (!result)
      result= new DOFVector<WorldVector<double> >(feSpace, "grad");

    WorldVector<double> grd;
    nullify(grd);

    result->set(grd);

    DOFVector<double> volume(feSpace, "volume");
    volume.set(0.0);

    DOFVector<WorldVector<double> > nrm(feSpace, "normal");

    const BasisFunction *basFcts = feSpace->getBasisFcts();
    int nBasisFcts = basFcts->getNumber();
    DimVec<double> bary(dim, DEFAULT_VALUE, (1.0 / (dim + 1.0)));

    Mesh *mesh = feSpace->getMesh();
    Flag flags(Mesh::CALL_LEAF_EL | Mesh::FILL_DET | Mesh::FILL_COORDS |
	      Mesh::FILL_GRD_LAMBDA | Mesh::FILL_BOUND);
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, flags);

    mtl::dense_vector<double> localUh(nBasisFcts);
    std::vector<DegreeOfFreedom> localIndices(nBasisFcts);
    std::map<DegreeOfFreedom, bool> visited;

    while (elInfo) {
      double det = elInfo->getDet();
      const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
      u->getLocalVector(elInfo->getElement(), localUh);
      basFcts->evalGrdUh(bary, grdLambda, localUh, grd);
      basFcts->getLocalIndices(elInfo->getElement(), feSpace->getAdmin(), localIndices);

      for (int i = 0; i < (dim + 1); i++) {
	  (*result)[localIndices[i]] += grd * det;
	  volume[localIndices[i]] += det;

	  if (elInfo->getBoundary(i)) {
	    WorldVector<double> n, rgrd;
	    elInfo->getNormal(i, n);

	    if (pointSym)
	      rgrd = grd;
	    else
	      rgrd = grd - 2.0*n*(grd*n); // reflect gradient at boundary

	    for (int j = 0; j < (dim + 1); j++)
	      if (j != i) {
		(*result)[localIndices[j]] += rgrd * det; // elements with boundary face
		volume[localIndices[j]] += det;

		nrm[localIndices[j]] += n;        // normal for other elements sharing boundary nodes
		visited[localIndices[j]] = true;
	      }
	  }
      }
      elInfo = stack.traverseNext(elInfo);
    }

    elInfo = stack.traverseFirst(mesh, -1, flags);
    while (elInfo) {
      double det = elInfo->getDet();
      const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
      basFcts->getLocalIndices(elInfo->getElement(), feSpace->getAdmin(), localIndices);

      for (int i = 0; i < (dim + 1); i++) {
	if (visited[localIndices[i]]) {
	  int j = 0;
	  for (; j < (dim + 1); j++)
	    if (j != i && elInfo->getBoundary(j))
	      break;

	  if (j == (dim + 1)) {
	    WorldVector<double> rgrd, n;
	    u->getLocalVector(elInfo->getElement(), localUh);
	    basFcts->evalGrdUh(bary, grdLambda, localUh, grd);

	    if (pointSym)
	      rgrd= grd;
	    else {
	      n = 1.0/norm(nrm[localIndices[i]]) * nrm[localIndices[i]];
	      rgrd = grd - 2.0*n*(grd*n);
	    }

	    (*result)[localIndices[i]] += rgrd * det;    // elements with boundary node but no face
	    volume[localIndices[i]] += det;
	  }
	}
      }
      elInfo = stack.traverseNext(elInfo);
    }

    DOFIterator<double> volIt(&volume, USED_DOFS);
    DOFIterator<WorldVector<double> > grdIt(result, USED_DOFS);

    for (volIt.reset(), grdIt.reset(); !volIt.end(); ++volIt, ++grdIt)
      if (*volIt != 0.0)
	*grdIt *= 1.0 / (*volIt);
  }

} }

#endif // EXTENSIONS_GRADIENT_CALCULATIONS_H
