/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#ifndef EXTENSIONS_VECTOR_OPERATIONS_H
#define EXTENSIONS_VECTOR_OPERATIONS_H

// #include "AMDiS.h"
#include "ValueTypes.h"

#if HAVE_PARALLEL_DOMAIN_AMDIS
#include "parallel/StdMpi.h"
#endif

#include <ios>
#include <iostream>
#include <fstream>
#include <string>
#include <boost/numeric/mtl/mtl.hpp>
#include <boost/numeric/itl/itl.hpp>
#include <boost/utility/enable_if.hpp>
#include <boost/type_traits.hpp>

#include "traits/size.hpp"
#include "traits/resize.hpp"
#include "traits/at.hpp"
#include "traits/num_cols.hpp"
#include "traits/num_rows.hpp"

namespace AMDiS { namespace extensions {

template<typename T>
const WorldVector<T>& operator*=(WorldVector<T>& v1, const WorldVector<T>& v2)
{
  T* v1It;
  T const* v2It;
  for (v1It = v1.begin(), v2It = v2.begin();
	  v1It != v1.end(); 
	  v1It++, v2It++)
  *v1It *= *v2It;
  
  return v1;
}

// x ~= y
template<typename T>
class compareTol : public std::binary_function<T, T, bool> {
public:
	compareTol(double tol_) : tol(tol_) {}
	bool operator()(const T &x, const T &y) const { return std::sqrt((x-y)*(x-y))<tol; }
	double tol;
};

template<typename T1, typename T2, int component>
struct comparePair : public std::binary_function<std::pair<T1, T2>, std::pair<T1, T2>, bool> 
{
  bool operator()(const std::pair<T1, T2> &x, const std::pair<T1, T2> &y) const 
  { 
    return (component==0?(x.first<y.first):(x.second<y.second));
  }
};

template<typename T1, typename T2>
struct compare0 : public std::binary_function<std::pair<T1, T2>, std::pair<T1, T2>, bool> 
{
  bool operator()(const std::pair<T1, T2> &x, const std::pair<T1, T2> &y) const 
  { 
    return x.first<y.first;
  }
};

template<typename T1, typename T2>
struct compare1 : public std::binary_function<std::pair<T1, T2>, std::pair<T1, T2>, bool> 
{
  bool operator()(const std::pair<T1, T2> &x, const std::pair<T1, T2> &y) const 
  { 
    return x.second<y.second;
  }
};

class SortPair : public std::binary_function<std::pair<WorldVector<double>, double>, std::pair<WorldVector<double>, double>, bool> {
public:
	SortPair(int component_=1,int idx_=0) : component(component_),idx(idx_) {}
	bool operator()(const std::pair<WorldVector<double>, double> &x, 
		const std::pair<WorldVector<double>, double> &y) const { 
		return (component==1?(x.first)[idx]<(y.first)[idx]:x.second<y.second); }
	int component,idx;
};

namespace vector_operations {
  
  // __________________________________________________________________________
  
  /// 2-norm for vectors
  template<typename Vector>
  typename boost::enable_if< is_vector<Vector>, typename ValueType<Vector>::type >::type
  norm(const Vector &b)
  {
    typename ValueType<Vector>::type value;
    nullify(value);
    for (size_t i = 0; i < num_rows(b); ++i)
      value += sqr(b[i]);
    return std::sqrt(value);
  }
  
  /// 2-norm for MTL4 vectors
  template<typename T>
  T norm(const mtl::dense_vector<T> &b) { return two_norm(b); }
  
  /// 1-norm for matrices
  template<typename Matrix>
  typename boost::enable_if< is_matrix<Matrix>, double >::type
  matrix_norm(const Matrix &M)
  {
    typename Matrix::value_type asum;
    asum = 0.0;
    for (size_t i = 0; i < num_rows(M); ++i) {
      for (size_t j = 0; j < num_cols(M); ++j)
	asum += std::abs(M[i][j]);
    }
    return static_cast<double>(asum);
  }
  
  
  /// 2-norm for vectors with eps-pertubation
  template<typename Vector>
  typename boost::enable_if< is_vector<Vector>, double >::type
  norm(const Vector &b, double eps)
  {
    return std::max(static_cast<double>(norm(b)), eps);
  }
  

  /// trace of a matrix
  template<typename Matrix>
  typename boost::enable_if< is_matrix<Matrix>, double >::type
  trace(Matrix &M)
  {
    TEST_EXIT_DBG(num_rows(M) == num_cols(M))
      ("Matrix dimensions must be equal!\n");
    typename Matrix::value_type sum;
    nullify(sum);
    for (size_t i = 0; i < num_rows(M); ++i)
      sum += M[i][i];
    return static_cast<double>(sum);
  }
  
  
  /// equivalent to matlab .* for vectors
  template<typename ResultVector, typename Vector1, typename Vector2>
  typename boost::enable_if< boost::mpl::and_< is_vector<Vector1>, is_vector<Vector2> >, ResultVector >::type
  mult(const Vector1& v1, const Vector2& v2)
  {
    TEST_EXIT_DBG(num_rows(v1)==num_rows(v2))
      ("Vectors must have the same size!\n");
    ResultVector result(num_rows(v1));
    for (size_t i = 0; i < num_rows(v1); ++i)
      result[i] = v1[i] * v2[i];
    return result;
  }
  
  
  template<typename ResultVector, typename Matrix, typename Vector>
  typename boost::enable_if< boost::mpl::and_< is_matrix<Matrix>, is_vector<Vector> >, ResultVector >::type
  mv(const Matrix &A, const Vector &x)
  {
    TEST_EXIT_DBG(num_cols(A)==num_rows(x))
      ("Matrix and vector dimension must fit together!");
    ResultVector result(num_rows(A));
    for (size_t i = 0; i < num_rows(A); ++i) {
      nullify(result[i]);
      for (size_t j = 0; j < num_cols(A); ++j)
	result[i] += at(A,i,j) * x[j];
    }
    return result;
  }
  
  
  template<typename ResultVector, typename Matrix, typename Vector1, typename Vector2>
  typename boost::enable_if< boost::mpl::and_< is_matrix<Matrix>, is_vector<Vector1>, is_vector<Vector2> >, 
			     ResultVector >::type
  Axpy(const Matrix &A, const Vector1 &x, const Vector2 &y)
  {
    TEST_EXIT_DBG(num_cols(A)==num_row(x) && num_rows(A)==num_rows(y))
      ("Matrix and vector dimensions must fit together!");
    ResultVector result;
    resize(result, num_rows(A));
    for (size_t i = 0; i < num_rows(A); ++i) {
      result[i] = y[i];
      for (size_t j = 0; j < num_cols(A); ++j)
	result[i] += at(A,i,j) * x[j];
    }
    return result;
  }

  // copy v -> w
  template<typename VectorIn, typename VectorOut>
  typename boost::enable_if< boost::mpl::and_< is_vector<VectorIn>, is_vector<VectorOut> >, void >::type
  copy(const VectorIn &v, VectorOut &w)
  {
    size_t s1 = size(v);
    size_t s2 = size(w);
    for (size_t i = 0; i < std::min(s1, s2); i++)
      w[i] = v[i];
  }
  
  template<class MatrixIn, class MatrixOut>
  typename boost::enable_if< boost::mpl::and_< is_matrix<MatrixIn>, is_matrix<MatrixOut> >, void >::type
  fillMatrix(const MatrixIn &m_in, MatrixOut &m)
  {
    resize(m, num_rows(m_in), num_cols(m_in));
    for (size_t i = 0; i < AMDiS::num_rows(m_in); ++i) {
      for (size_t j = 0; j < AMDiS::num_cols(m_in); ++j) {
	at(m,i,j) = at(m_in,i,j);
      }
    }
  }
  
  template<class MatrixIn, class MatrixOut>
  typename boost::enable_if< boost::mpl::and_< is_matrix<MatrixIn>, is_matrix<MatrixOut> >, void >::type
  copy(const MatrixIn &m_in, MatrixOut &m)
  {
    fillMatrix(m_in, m);
  }
  
  // v3:=[v1, v2]
  template<typename Vector1, typename Vector2, typename VectorOut>
  typename boost::enable_if< boost::mpl::and_< is_vector<Vector1>, is_vector<Vector2>, is_vector<VectorOut> >, 
			     void >::type
  merge(Vector1 &v1, Vector2 &v2, VectorOut &v3)
  {
    resize(v3, num_rows(v1) + num_rows(v2));
    for (size_t i = 0; i < num_rows(v1); i++)
      v3[i] = v1[i];
    for (size_t j = 0; j < num_rows(v2); j++)
      v3[j + num_rows(v1)] = v2[j];
  }
  
  template<typename Vector>
  void getMin(const Vector &v, typename ValueType<Vector>::type &minVal, size_t &minIdx)
  {
    TEST_EXIT(num_rows(v) > 0)("getMin of empty vector!\n");
    
    minVal = v[0];
    minIdx = 0;
    for (size_t i = 1; i < num_rows(v); i++) {
      if (v[i] < minVal) {
	minVal = v[i];
	minIdx = i;
      }
    }
  }
  
  template<typename Vector>
  void getMax(const Vector &v, typename ValueType<Vector>::type &maxVal, size_t &maxIdx)
  {
    TEST_EXIT(num_rows(v) > 0)("getMax of empty vector!\n");
    
    maxVal = v[0];
    maxIdx = 0;
    for (size_t i = 1; i < num_rows(v); i++) {
      if (v[i] > maxVal) {
	maxVal = v[i];
	maxIdx = i;
      }
    }
  }

  template<typename Vector>
  typename ValueType<Vector>::type
  min(const Vector &vec)
  {
    typename ValueType<Vector>::type minVal;
    size_t minIdx;
    getMin(vec, minVal, minIdx);
    return minVal;
  }

  template<typename Vector>
  typename ValueType<Vector>::type
  max(const Vector &vec)
  {
    typename ValueType<Vector>::type maxVal;
    size_t maxIdx;
    getMax(vec, maxVal, maxIdx);
    return maxVal;
  }
  
  template<typename Vector>
  typename ValueType<Vector>::type
  absmin(const Vector &v)
  {
    typedef typename ValueType<Vector>::type T;
    TEST_EXIT(num_rows(v) > 0)("absmin of empty vector!\n");
    
    T minVal = abs(v[0]);
    for(size_t i = 1; i < num_rows(v); i++) {
      if (std::abs(v[i]) < minVal) {
	minVal = std::abs(v[i]);
      }
    }
    return minVal;
  }
  
  template<typename Vector>
  typename ValueType<Vector>::type
  absmax(const Vector &v)
  {
    typedef typename ValueType<Vector>::type T;
    TEST_EXIT(num_rows(v) > 0)("absmax of empty vector!\n");
    
    T maxVal = abs(v[0]);
    for(size_t i = 1; i < num_rows(v); i++) {
      if (std::abs(v[i]) > maxVal) {
	maxVal = std::abs(v[i]);
      }
    }
    return maxVal;
  }
  
  template<typename Vector>
  typename ValueType<Vector>::type
  sum(const Vector &v)
  {
    typedef typename ValueType<Vector>::type T;
    size_t N = num_rows(v);
    T value; nullify(value);
    if (N > 0) {
      value = v[0];
      for (size_t i = 1; i < N; ++i)
	value += v[i];
    }
    return value;
  }
  
  template<typename Vector>
  inline typename ValueType<Vector>::type
  mean(const Vector &v)
  {
    double N = static_cast<double>(num_rows(v));
    return sum(v) * (1.0 / N);
  }
  
  template<typename Vector>
  void getMean(const Vector &v, typename ValueType<Vector>::type &meanValue)
  {
    meanValue = mean(v);
  }
  
  template<typename T1, typename T2, typename Compare>
  void sort(std::vector<T1> &vec1, std::vector<T2> &vec2, Compare &comp)
  {
    std::vector<std::pair<T1,T2> > order(vec1.size());
    
    // create vector of pairs
    for (size_t i=0; i<vec1.size(); ++i)
      order[i] = std::make_pair(vec1[i], vec2[i]);
    
    // sort vector of pairs
    std::sort(order.begin(), order.end(), comp);
    
    // copy sorted pairs back to vectors
    for (size_t i=0; i<vec1.size(); ++i) {
      vec1[i] = order[i].first;
      vec2[i] = order[i].second;
    }
  }
  
  template<typename T1, typename T2, typename Compare>
  void sort(std::vector<std::pair<T1,T2> > &vec, Compare &comp)
  {
    std::sort(vec.begin(), vec.end(), comp);
  }
  
  inline void unique(std::vector<WorldVector<double> > &vec, double tol, std::vector<unsigned> &ind)
  {
    compareTol<WorldVector<double> > comp(tol);
    size_t newVec = 0;
    for (size_t i = 0; i < vec.size(); ++i) {
      bool inNew = false;
      for (size_t j = 0; j < newVec; ++j) {
	inNew = inNew || comp(vec[i], vec[j]);
	if(inNew) 
	  break;
      }
      if (!inNew) {
	vec[newVec] = vec[i]; 
	newVec++; 
	ind.push_back(i);
      }
    }
    vec.erase(vec.begin()+newVec,vec.end());
  }
  
  template<typename T>
  inline void filter(std::vector<T> &x, const std::vector<bool> &filter)
  {
    TEST_EXIT_DBG(num_rows(x) == num_rows(filter))
      ("number of rows of x and filter must be equal!\n");
    size_t j = 0;
    for (size_t i = 0; i < num_rows(x); ++i) {
      if (filter[i]) {
	x[j] = x[i];
	j++;
      }
    }
    x.erase(x.begin()+j,x.end());
  }
  
  template<typename T>
  void filter(std::vector<T> &x, const std::vector<int> &filter)
  {
    std::vector<T> x_(num_rows(filter));
    for (size_t i = 0; i < num_rows(filter); ++i)
      x_[i] = x[filter[i]];
    swap(x, x_);
  }
 
  template<typename Vector1, typename Vector2>
  typename AMDiS::traits::mult_type<typename ValueType<Vector1>::type, typename ValueType<Vector2>::type>::type
  dot(const Vector1 &vec1, const Vector2 &vec2)
  {
    typedef typename AMDiS::traits::mult_type<typename ValueType<Vector1>::type, typename ValueType<Vector2>::type>::type value_type;
    value_type value; nullify(value);
    
    size_t N = num_rows(vec1);
    if (N > 0) {
      value = vec1[0] * vec2[0];
      for (size_t i = 1; i < N; ++i)
	value += vec1[i] * vec2[i];
    }
    return value;
  }
}

} }

#endif // EXTENSIONS_VECTOR_OPERATIONS_H
