/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#include "Expressions.h"

namespace AMDiS { namespace base_problems {

namespace detail {

template<typename P>
NavierStokes_TH_MultiPhase<P>::NavierStokes_TH_MultiPhase(const std::string &name_, bool createProblem) 
  : super(name_, createProblem),
    viscosity1(1.0),
    viscosity2(1.0),
    density1(1.0),
    density2(1.0),
    phaseFieldType(0),
    densityPhase(NULL),
    viscosityPhase(NULL),
    multiPhase(NULL),
    oldSolution(NULL),
    stepSolution(NULL)
{ FUNCNAME("NavierStokes_TH_MultiPhase::_constructor()");

  Initfile::get(self::name + "->viscosity1", viscosity1); // viscosity of fluid 1
  Initfile::get(self::name + "->viscosity2", viscosity2); // viscosity of fluid 2
  Initfile::get(self::name + "->density1", density1); // density of fluid 1
  Initfile::get(self::name + "->density2", density2); // density of fluid 2  

  phaseFieldType = 1;  
  Initfile::get(self::name + "->phase field type", phaseFieldType); // 0: [0,1], 1: [-1,1]

  if (viscosity1 <= 0.0 && self::viscosity > 0.0)
    viscosity1 = self::viscosity;
  
  TEST_EXIT(viscosity1 > 0.0)("Positive viscosity is necessary!\n");
}


template<typename P>
NavierStokes_TH_MultiPhase<P>::~NavierStokes_TH_MultiPhase()
{
  if (densityPhase) { 
    delete densityPhase; 
    densityPhase = NULL; 
  }
  if (viscosityPhase) { 
    delete viscosityPhase; 
    viscosityPhase = NULL; 
  }
  if (oldSolution) {
    delete oldSolution;
    oldSolution = NULL;
  }
  if (stepSolution) {
    delete stepSolution;
    stepSolution = NULL;
  }
}


template<typename P>
void NavierStokes_TH_MultiPhase<P>::initData()
{
  densityPhase = new DOFVector<double>(self::prob->getFeSpace(0), "densityPhase");
  viscosityPhase = new DOFVector<double>(self::prob->getFeSpace(0), "viscosityPhase");
  
  oldSolution = new SystemVector(*self::prob->getSolution());
  stepSolution = new SystemVector(*self::prob->getSolution());
  
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    Parallel::MeshDistributor::globalMeshDistributor->addInterchangeVector(densityPhase);
    Parallel::MeshDistributor::globalMeshDistributor->addInterchangeVector(viscosityPhase);
#endif

  densityPhase->set(density1);
  viscosityPhase->set(viscosity1);

  super::initData();
}


template<typename P>
void NavierStokes_TH_MultiPhase<P>::transferInitialSolution(AdaptInfo *adaptInfo)
{
  super::transferInitialSolution(adaptInfo);
  *oldSolution = *self::prob->getSolution();
  *stepSolution = *self::prob->getSolution();
}


template<typename P>
void NavierStokes_TH_MultiPhase<P>::initTimestep(AdaptInfo *adaptInfo)
{
  super::initTimestep(adaptInfo);
  if (multiPhase) {
    if (phaseFieldType == 0) {
      *viscosityPhase << viscosity2 + valueOf(multiPhase)*(viscosity1 - viscosity2);
      *densityPhase << density2 + valueOf(multiPhase)*(density1 - density2);
    } else {
      *viscosityPhase << function_(LinearInterpolation1(viscosity1, viscosity2), valueOf(multiPhase));
      *densityPhase << function_(LinearInterpolation1(density1, density2), valueOf(multiPhase));
    }
  }
}


template<typename P>
void NavierStokes_TH_MultiPhase<P>::closeTimestep(AdaptInfo *adaptInfo)
{
  super::closeTimestep(adaptInfo);
  acceptTimestep();
}


template<typename P>
void NavierStokes_TH_MultiPhase<P>::fillOperators()
{
  using namespace extensions;
  
  WorldVector<DOFVector<double>* > vel, vel_old;
  for (unsigned i = 0; i < self::dow; i++) {
    vel[i]= stepSolution->getDOFVector(i);
    vel_old[i]= oldSolution->getDOFVector(i);
  }
  
  // fill operators for prob
  for (size_t i = 0; i < self::dow; ++i) {
    /// < (1/tau)*rho*u'_i , psi >
    Operator *opTime = new Operator(self::getFeSpace(i), self::getFeSpace(i));
    if (density1 == density2 || !densityPhase)
      addZOT(opTime, density1);
    else
      addZOT(opTime, valueOf(densityPhase));
    self::prob->addMatrixOperator(*opTime, i, i, self::getInvTau(), self::getInvTau());
    
    /// < (1/tau)*rho*u_i^old , psi >
    Operator *opTimeOld = new Operator(self::getFeSpace(i), self::getFeSpace(i));
    if (density1 == density2 || !densityPhase)
      addZOT(opTimeOld, density1 * valueOf(vel_old[i]));
    else
      addZOT(opTimeOld, valueOf(densityPhase) * valueOf(vel_old[i]));
    self::prob->addVectorOperator(*opTimeOld, i, self::getInvTau(), self::getInvTau());
 
    /// < u^old*grad(u_i^old) , psi >
    Operator *opUGradU0 = new Operator(self::getFeSpace(i),self::getFeSpace(i));
    if (density1 == density2 || !densityPhase)
      addFOT(opUGradU0, density1 * valueOf(vel), GRD_PHI);
    else
      addFOT(opUGradU0, valueOf(densityPhase) * valueOf(vel), GRD_PHI);
    opUGradU0->setUhOld(vel[i]);
    self::prob->addVectorOperator(*opUGradU0, i);


    /// < u'*grad(u_i^old) , psi >
    for (size_t j = 0; j < self::dow; ++j) {
      Operator *opUGradU1 = new Operator(self::getFeSpace(i),self::getFeSpace(j));
      if (density1 == density2 || !densityPhase)
	addZOT(opUGradU1, density1 * derivativeOf(vel[i], j));
      else
	addZOT(opUGradU1, valueOf(densityPhase) * derivativeOf(vel[i], j));
      self::prob->addMatrixOperator(*opUGradU1, i, j);
    }
      
    /// < u^old*grad(u'_i) , psi >
    for (size_t j = 0; j < self::dow; ++j) {
      Operator *opUGradU2 = new Operator(self::getFeSpace(i),self::getFeSpace(i));
      if (density1 == density2 || !densityPhase)
	addFOT(opUGradU2, density1 * valueOf(vel[j]), j, GRD_PHI);
      else
	addFOT(opUGradU2, valueOf(densityPhase) * valueOf(vel[j]), j, GRD_PHI);
      self::prob->addMatrixOperator(*opUGradU2, i, i);
    }

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity
    addLaplaceTerm(i);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(self::getFeSpace(i),self::getFeSpace(self::dow));
    addFOT(opGradP, -1.0, i, GRD_PSI);
    self::prob->addMatrixOperator(*opGradP, i, self::dow);
    
    /// external force, e.g. gravitational force
    if (std::abs(self::force[i]) > DBL_TOL) {
      Operator *opForce = new Operator(self::getFeSpace(i), self::getFeSpace(i));
      if (density1 == density2 || !densityPhase)
	addZOT(opForce, density1*self::force[i] );
      else
	addZOT(opForce, valueOf(densityPhase)*self::force[i] );
      self::prob->addVectorOperator(*opForce, i);
    }
  }

  /// div(u) = 0
  for (size_t i = 0; i < self::dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(self::getFeSpace(self::dow),self::getFeSpace(i));
    addFOT(opDivU, 1.0, i, GRD_PHI);
    self::prob->addMatrixOperator(*opDivU, self::dow, i);
  }
  
  Operator *opZero = new Operator(self::getFeSpace(self::dow),self::getFeSpace(self::dow));
  self::prob->addMatrixOperator(*opZero, self::dow, self::dow);
}


template<typename P>
void NavierStokes_TH_MultiPhase<P>::addLaplaceTerm(int i)
{
  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (viscosity1 != viscosity2 && viscosityPhase) {
    for (unsigned j = 0; j < self::dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(self::getFeSpace(i), self::getFeSpace(j));
      addSOT(opLaplaceUi1, valueOf(viscosityPhase), j, i);
      self::prob->addMatrixOperator(*opLaplaceUi1, i, j);
    }
  }
  
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(self::getFeSpace(i), self::getFeSpace(i));
  if (viscosity1 == viscosity2 || !viscosityPhase)
    addSOT(opLaplaceUi, viscosity1);
  else
    addSOT(opLaplaceUi, valueOf(viscosityPhase));
  self::prob->addMatrixOperator(*opLaplaceUi, i, i);
}

} // end namespace detail

} }

