/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef NAVIER_STOKES_TAYLOR_HOOD_H
#define NAVIER_STOKES_TAYLOR_HOOD_H

#include "AMDiS.h"
#include "BaseProblem.h"
#include "ExtendedProblemStat.h"
#include "Expressions.h"

namespace AMDiS { namespace base_problems {

namespace detail 
{  
  /** \ingroup NavierStokes_TaylorHood
  * \brief
  * Navier-Stokes problem, using Taylor Hood elements
  * and theta-scheme time-integration
  */
  
  template<typename ProblemStatType>
  class NavierStokes_TaylorHood : public BaseProblem<ProblemStatType>
  {
  public: // typedefs
    
    typedef NavierStokes_TaylorHood<ProblemStatType> self;
    typedef BaseProblem<ProblemStatType> super;
    
  public: // methods

    NavierStokes_TaylorHood(const std::string &name_, bool createProblem = true);

    virtual ~NavierStokes_TaylorHood();

    virtual void initData() override;

    virtual void transferInitialSolution(AdaptInfo *adaptInfo) override;  

    virtual void closeTimestep(AdaptInfo *adaptInfo) override;
    
    virtual void writeFiles(AdaptInfo *adaptInfo, bool force = false) override;

    // === getting/setting methods ===

    DOFVector<WorldVector<double> >* getVelocity()
    { FUNCNAME_DBG("getVelocity()");
      TEST_EXIT_DBG(velocity != NULL)
	("velocity is NULL!");
      return velocity;
    }
    
    DOFVector<double> *getOldSolution(int i)
    { FUNCNAME_DBG("getOldSolution()");
      TEST_EXIT_DBG(oldSolution[i] != NULL)
	("Index out of range, or oldSolution not provided for this component!\n");
      return oldSolution[i];
    }
    
    double getDensity() { return density; }
    double getViscosity() { return viscosity; }

    virtual void fillOperators() override;
    
    virtual void addLaplaceTerm(int i);

  protected: // variables

    DOFVector<WorldVector<double> >* velocity;

    void calcVelocity()
    {
      WorldVector<DOFVector<double>*> vel;
      for (size_t i = 0; i < self::dow; i++)
	vel[i] = self::prob->getSolution()->getDOFVector(i);
      *velocity << valueOf(vel);
    }

    int laplaceType;
    int nonLinTerm;

    double oldTimestep;
    double viscosity;
    double density;
    
    double minus1;
    double theta;
    double theta1;
    double minusTheta1;

    WorldVector<double> force;

    FileVectorWriter *fileWriter;
    std::vector<DOFVector<double>*> oldSolution;
  };

} // end namespace detail

typedef detail::NavierStokes_TaylorHood<AMDiS::ProblemStat> NavierStokes_TaylorHood;

} }


#include "NavierStokes_TaylorHood.hh"

#endif // NAVIER_STOKES_TAYLOR_HOOD_H

