#ifndef MATRIX_SHELL_CH_HPP
#define MATRIX_SHELL_CH_HPP

#include "solver/BlockMTLMatrix.h"
#include "solver/itl/block_diagonal.hpp"

namespace AMDiS { namespace extensions {

using namespace AMDiS::MTLTypes;

/// Cahn-Hilliard equation as matrix-shell
struct MTLMatrixShellCahnHilliard : BlockMTLMatrix
{
  typedef BlockMTLMatrix                                   super;
  typedef DOFMatrix::base_matrix_type                      MatrixType;
  typedef typename mtl::Collection<MatrixType>::size_type  size_type;
  
  MTLMatrixShellCahnHilliard() {}
  
  template<typename Data>
  void setData(Data* data)
  {
    tau = data->tau;
    eps = data->eps;
    gamma = data->gamma;
    // provide some matrices
    matM = &data->getM();
    matL = &data->getL();
    matN = &data->getN();
    matK = &data->getK();
  }
  
  const MatrixType& getM() const { return *matM; } // c
  const MatrixType& getL() const { return *matL; } // laplace(c)
  const MatrixType& getN() const { return *matN; } // d_c(Phi'(c))
  const MatrixType& getK() const { return *matK; } // div(M(c) * grad(mu))
  
  /// return delta = \ref tau * \ref gamma
  double getDelta() const { return (*tau) * gamma; }

  /// Evaluate the matrix-vector product x := A*b
  /** With matrix A defined as
   *
   * A  = [ M           , M + N - eps^2*L ]
   *      [ tau*gamma*K , M               ]
   **/
  template <typename VectorIn, typename VectorOut, typename Assign>
  void mult(const VectorIn& b, VectorOut& x, Assign) const
  {
    TEST_EXIT_DBG(initialized)("Block matrix not initialized. Assign a SolverMatrix first!\n");
      
    // Cahn-Hilliard equation
    VectorOut y0(x[getRowRange(0)]);
    VectorOut y1(x[getRowRange(1)]);
      
    const VectorIn c0(b[getRowRange(0)]);
    const VectorIn c1(b[getRowRange(1)]);
      
    // y0 = M*c0 + M*c1 + N*c1 - eps^2*L*c1
    Assign::first_update(y0, getM() * (c0 + c1) );
    Assign::update(      y0, getL() * ((-sqr(eps))*c1) );
    Assign::update(      y0, getN() * c1 );
      
    // y1 = tau*gamma*K*c0 + M*c1
    Assign::first_update(y1, getK() * (getDelta() * c0) );
    Assign::update(      y1, getM() * c1 );
  }

  template <typename VectorIn>
  mtl::vector::mat_cvec_multiplier<MTLMatrixShellCahnHilliard, VectorIn> operator*(const VectorIn& v) const
  {   return mtl::vector::mat_cvec_multiplier<MTLMatrixShellCahnHilliard, VectorIn>(*this, v);    }

private:  
  MTLMatrix* matM;
  MTLMatrix* matL;
  MTLMatrix* matN;
  MTLMatrix* matK;
  
  double eps;
  double gamma;
  double* tau;
};

} } // end namespaces

// some necessary traits
// _____________________________________________________________________________


inline AMDiS::extensions::MTLMatrixShellCahnHilliard::size_type size(const AMDiS::extensions::MTLMatrixShellCahnHilliard& A) { return (A.n) * (A.m); }
inline AMDiS::extensions::MTLMatrixShellCahnHilliard::size_type num_rows(const AMDiS::extensions::MTLMatrixShellCahnHilliard& A) { return A.n; }
inline AMDiS::extensions::MTLMatrixShellCahnHilliard::size_type num_cols(const AMDiS::extensions::MTLMatrixShellCahnHilliard& A) { return A.m; }

namespace mtl 
{
  template <>
  struct Collection<AMDiS::extensions::MTLMatrixShellCahnHilliard>
  {
    typedef double value_type;
    typedef int    size_type;
  };

  namespace ashape 
  {
    template <> struct ashape_aux<AMDiS::extensions::MTLMatrixShellCahnHilliard> 
    { 
      typedef nonscal type;
    };
  } // end namespace ashape  
} // end namespace mtl


namespace AMDiS 
{
  namespace traits 
  {
    /// size
    template <>
    struct size< AMDiS::extensions::MTLMatrixShellCahnHilliard >
    {
	typedef AMDiS::extensions::MTLMatrixShellCahnHilliard::size_type   type;
	type operator()(const AMDiS::extensions::MTLMatrixShellCahnHilliard& A) const 
	{ return A.n * A.m; }
    };
    
    /// num rows
    template <>
    struct num_rows< AMDiS::extensions::MTLMatrixShellCahnHilliard >
    {
	typedef AMDiS::extensions::MTLMatrixShellCahnHilliard::size_type   type;
	type operator()(const AMDiS::extensions::MTLMatrixShellCahnHilliard& A) const 
	{ return A.n; }
    };
    
    /// num cols
    template <>
    struct num_cols< AMDiS::extensions::MTLMatrixShellCahnHilliard >
    {
	typedef AMDiS::extensions::MTLMatrixShellCahnHilliard::size_type   type;
	type operator()(const AMDiS::extensions::MTLMatrixShellCahnHilliard& A) const 
	{ return A.n; }
    };
  }

  template<>
  void CreatorMap<ITL_PreconditionerBase<extensions::MTLMatrixShellCahnHilliard, MTLTypes::MTLVector> >::addDefaultCreators()
  {
    typedef ITL_Preconditioner<itl::pc::identity<extensions::MTLMatrixShellCahnHilliard>, extensions::MTLMatrixShellCahnHilliard, MTLTypes::MTLVector > ChBlockIdentityPreconditioner;
    addCreator("no", new ChBlockIdentityPreconditioner::Creator);
  }
  
  template< typename Mapper >
  inline void operator<<(extensions::MTLMatrixShellCahnHilliard& mat, MatMap<const SolverMatrix<Matrix<DOFMatrix* > >, Mapper >& mapper)
  {
    mat.setMatrix(mapper.mat);
  }
} // end namespaces AMDiS

#endif // MATRIX_SHELL_CH_HPP