/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#include "AMDiS.h"
#include "base_problems/CahnHilliard.h"
#include "Expressions.h"
#include "Helpers.h"
#include "RefinementExpression.h"

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
#include "parallel/PetscSolverCahnHilliard.h"
#else
#include "preconditioner/MTLPreconCahnHilliard.h"
#include "preconditioner/MTLPreconCahnHilliard2.h"
#endif

using namespace AMDiS;

// data used by the preconditioner
struct DataContainer
{
  typedef typename DOFMatrix::base_matrix_type MatrixType;

  DataContainer(std::string name, size_t N)
    : name(name),
      matM(NULL), matL(NULL),
      tau(NULL), eps(0.1), gamma(1.0)
  {
    Parameters::get(name + "->epsilon", eps);
    Parameters::get(name + "->gamma", gamma);
  }
  
  MatrixType& getM() { return matM->getBaseMatrix(); }
  MatrixType& getL() { return matL->getBaseMatrix(); }
  
  std::string name;
  
  DOFMatrix* matM;
  DOFMatrix* matL;
  
  double* tau;
  double eps;
  double gamma;
};

class MyCahnHilliard : public CahnHilliard
{
public:
  typedef CahnHilliard super;
  
  MyCahnHilliard(std::string name)
    : super(name), 
      data(NULL), 
      matrixL(NULL), 
      matrixM(NULL), 
      refineFct(NULL),
      mesh_changed(true)
  { }
  
  ~MyCahnHilliard() 
  {
    if (matrixL) {
      delete matrixL;
      matrixL = NULL;
    }
    if (matrixM) {
      delete matrixM;
      matrixM = NULL;
    }
    if (data) {
      delete data;
      data = NULL;
    }
    if (refineFct) {
      delete refineFct;
      refineFct = NULL;
    }
  }
  
  void initData()
  {
    super::initData();
    
    prob->getMesh()->dofCompress();
    matrixL = new DOFMatrix(prob->getFeSpace(), prob->getFeSpace());
    matrixM = new DOFMatrix(prob->getFeSpace(), prob->getFeSpace());
    
    data = new DataContainer("ch", prob->getFeSpace(0)->getAdmin()->getUsedSize());
    data->tau = getTau();
    data->eps = eps;
    data->matM = matrixM;
    data->matL = matrixL;
    
    refineFct = new PhaseFieldRefineFct("refinement->phasefield", prob->getMesh());
  }
    
  void finalizeData()
  {
    super::finalizeData();
    
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    Parallel::PetscSolverCahnHilliard* solver = dynamic_cast<Parallel::PetscSolverCahnHilliard*>(prob->getSolver());
    if (solver)
      solver->setChData2(&eps, &gamma, getTau());
#else
    extensions::MTLPreconCahnHilliard* precon = dynamic_cast<extensions::MTLPreconCahnHilliard*>(prob->getSolver()->getRightPrecon());
    if (precon)
      precon->setData(data);
    
    extensions::MTLPreconCahnHilliard2* precon2 = dynamic_cast<extensions::MTLPreconCahnHilliard2*>(prob->getSolver()->getRightPrecon());
    if (precon2)
      precon2->setData(data);
#endif
  }
  
  void fillBoundaryConditions() override
  {
    for (int i = 0; i < prob->getNumComponents(); i++) {
      for (int j = 0; j < prob->getNumComponents(); j++) {
	for (int k = 1; k <= Global::getGeo(WORLD); k++)
	  prob->addPeriodicBC(-k, i, j);
      }
    }
  }
  
  void solveInitialProblem(AdaptInfo *adaptInfo) override
  {
//     super::solveInitialProblem(adaptInfo);
    
    // random initial data
    // nach prob->getSolution(1) !!
    prob->getSolution(1)->interpol(new AMDiS::Random(0.0, 1.0));
  }
  
  void initTimestep(AdaptInfo* adaptInfo) override
  {
    super::initTimestep(adaptInfo);
    
    if (mesh_changed) {
      prob->getMesh()->dofCompress();
      Operator laplaceOp(prob->getFeSpace(), prob->getFeSpace());
      addSOT(laplaceOp, 1.0);
      matrixL->assembleOperator(laplaceOp);
	
      Operator massOp(prob->getFeSpace(), prob->getFeSpace());
      addZOT(massOp, 1.0);
      matrixM->assembleOperator(massOp);
    }
  }
  
  void closeTimestep(AdaptInfo* adaptInfo) override
  {
    DOFVector<double>* c = prob->getSolution(1);
    mesh_changed = RefinementExpression(*prob).refine( function_(*refineFct, valueOf(c)) );
    
    super::closeTimestep(adaptInfo);
  }
  
  DOFMatrix* getMatrixL() { return matrixL; }
  DOFMatrix* getMatrixM() { return matrixM; }
  
protected:
  DataContainer* data;
  
  DOFMatrix* matrixL;
  DOFMatrix* matrixM;
  
  PhaseFieldRefineFct* refineFct;
  bool mesh_changed;
};



int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);   
  Timer t;
  
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  CreatorMap< LinearSolverInterface >::addCreator("p_petsc_petsc-ch", new Parallel::PetscSolverCahnHilliard::Creator);
#else
  typedef typename extensions::MTLPreconCahnHilliard::precon_base CHBasePreconditioner;
  CreatorMap< CHBasePreconditioner >::addCreator("ch", new extensions::MTLPreconCahnHilliard::Creator);
  CreatorMap< CHBasePreconditioner >::addCreator("ch2", new extensions::MTLPreconCahnHilliard2::Creator);
#endif
  
  
  MyCahnHilliard chProb("ch");
  chProb.initialize(INIT_ALL);
  chProb.initTimeInterface();
    
  // scale the mesh to a given size
  bool scaleMesh = false;
  Initfile::get("mesh->scale mesh",scaleMesh);
  if (scaleMesh) {
    WorldVector<double> scale; scale.set(1.0);
    Initfile::get("mesh->dimension",scale);
    Helpers::scaleMesh(chProb.getMesh(), scale);
  }
  
  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", chProb.getNumComponents());
  AdaptInstationary adaptInstat("adapt", chProb, adaptInfo, chProb, adaptInfo);

  int error_code = adaptInstat.adapt(); 

  MSG("elapsed time= %d sec\n", t.elapsed());
  AMDiS::finalize();

  return error_code;
}
