#include "AMDiS.h"
#include "Helpers.h"

#include "MTLMatrixShellCahnHilliard.hpp"
#include "CahnHilliardProblemStat.hpp"

#include "preconditioner/MTLPreconCahnHilliard.h"
#include "solver/BITL_Solver.h"

using namespace boost::posix_time;
using namespace AMDiS;
  
typedef DOFMatrix::base_matrix_type BlockType;
typedef MTLMatrixShellCahnHilliard MatrixShell;
typedef MTLPreconChBase<MatrixShell, BlockType> Precon;

// data used by the preconditioner
struct DataContainer
{

  DataContainer(std::string name, size_t N)
    : name(name),
      matM(NULL), matL(NULL),
      tau(NULL), eps(0.1), gamma(1.0)
  {
    Parameters::get(name + "->epsilon", eps);
    Parameters::get(name + "->gamma", gamma);
  }
  
  BlockType& getM() { return matM->getBaseMatrix(); }
  BlockType& getL() { return matL->getBaseMatrix(); }
  BlockType& getN() { return matN->getBaseMatrix(); }
  BlockType& getK() { return matK->getBaseMatrix(); }
  
  std::string name;
  
  DOFMatrix* matM;
  DOFMatrix* matL;
  DOFMatrix* matN;
  DOFMatrix* matK;
  
  double* tau;
  double eps;
  double gamma;
};

template<typename ProblemStatType>
struct CahnHilliardInstat : ProblemInstat
{
  CahnHilliardInstat(std::string name, ProblemStatType& prob_)
    : ProblemInstat(name, prob_), prob(prob_), data(NULL) {}
    
  ~CahnHilliardInstat()
  {
    if (data) {
      delete data;
      data = NULL;
    }
  }
    
  void initData(AdaptInfo* adaptInfo)
  {
    // fill data container
    size_t N = prob.getFeSpace(0)->getAdmin()->getUsedSize();
    data = new DataContainer("ch", N);
    data->tau = adaptInfo->getTimestepPtr();
    Parameters::get(prob.getName() + "->eps", data->eps);
    Parameters::get(prob.getName() + "->gamma", data->gamma);
    
    // create the operators
    prob.setData(data);
    prob.fillOperators();
    
    // collect pointers to assembled matrices
    data->matM = prob.getSystemMatrix(0,0);
    data->matL = prob.getSystemMatrix(1,1);
    data->matN = prob.getSystemMatrix(1,0);
    data->matK = prob.getSystemMatrix(0,1);
      
    // push data to system matrix
    LinearSolverBase<MatrixShell>* solver = dynamic_cast<LinearSolverBase<MatrixShell>*>(prob.getSolver());
    if (solver)
      solver->getMatrix().setData(data);
    
    // push data to preconditioner
    Precon* precon = dynamic_cast<Precon*>(prob.getSolver()->getRightPrecon());
    if (precon)
      precon->setData(data);
  }
  
  void solveInitialProblem(AdaptInfo* adaptInfo)
  {
    prob.getSolution(1)->interpol(new AMDiS::Random(0.0, 1.0));
  }
  
protected:
  ProblemStatType& prob;
  DataContainer* data;
};

// -----------------------------------------------------------------------------

int main(int argc, char** argv)
{ FUNCNAME("pfc");
  
  AMDiS::init(argc, argv);
  
  using namespace AMDiS::MTLTypes;
  typedef ITL_PreconditionerBase<MatrixShell, MTLTypes::MTLVector> MatrixShellPreconditionerType;
  CreatorMap<MatrixShellPreconditionerType>::addCreator("ch", new Precon::Creator);
  
  initCreatorMap<MatrixShell>("cahn_hilliard");
  // backend: cahn_hilliard, solver: fgmres
  
  Timer t;
  
  CahnHilliardProblemStat prob("ch");
  prob.initialize(INIT_ALL);
  CahnHilliardInstat<CahnHilliardProblemStat> instat("ch", prob);
  instat.initialize(INIT_ALL);
  
  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", prob.getNumComponents());
  AdaptInstationary adaptInstat("adapt", prob, adaptInfo, instat, adaptInfo);
  
  // scale the mesh to a given size
  bool scaleMesh = false;
  Initfile::get("mesh->scale mesh",scaleMesh);
  if (scaleMesh) {
    WorldVector<double> scale; scale.set(1.0);
    Initfile::get("mesh->dimension",scale);
    Helpers::scaleMesh(prob.getMesh(), scale);
  }
    
  instat.initData(&adaptInfo);
  int error_code = adaptInstat.adapt(); 
  
  MSG("elapsed time= %d sec\n", t.elapsed());
  AMDiS::finalize();
  return 0;
}
