#include "AMDiS.h"
#include <fstream>
#include "Refinement.h"
#include "Refinement_Level.h"
#include "SignedDistFunctors.h"
#include "Helpers.h"
#include "PhaseFieldConvert.h"
#include "GenericOperatorTerm.h"
#include "ExtendedProblemStat.h"


#define USE_UH_OLD 1


using namespace AMDiS;


class MySignedDistRefinement : public SignedDistRefinement
{
public:
    MySignedDistRefinement(Mesh* mesh_) : SignedDistRefinement(mesh_) 
    {
      Parameters::get("mesh2->refinement->level in inner domain",lInner);
      Parameters::get("mesh2->refinement->level in outer domain",lOuter);
      Parameters::get("mesh2->refinement->level on interface",lInterface);

      lInterface-= mesh->getMacroElementLevel();
      lInner-= mesh->getMacroElementLevel();
      lOuter-= mesh->getMacroElementLevel();

      int local_globalSize = 10;
      Parameters::get("mesh2->refinement->initial level", local_globalSize);
      globalSize = local_globalSize;

      Parameters::get("mesh2->refinement->interface width",interfaceWidth);
      Parameters::get("mesh2->refinement->fade out width",fadeOutWidth);
      Parameters::get("mesh2->refinement->sign in inner domain",signInInnerDomain);
    };
};


/// Dirichlet boundary function
class G : public AbstractFunction<double, WorldVector<double> >,
	  public TimedObject
{
public:
  /// Implementation of AbstractFunction::operator().

  double operator()(const WorldVector<double>& x) const 
  {
    return 1.0;
  }
};



/// RHS function
class F : public AbstractFunction<double, WorldVector<double> >,
	  public TimedObject
{
public:
  F(int degree) : AbstractFunction<double, WorldVector<double> >(degree) {}
  
  /// Implementation of AbstractFunction::operator().
  double operator()(const WorldVector<double>& x) const 
  {
    return 1.0;
  }
};

struct DDProblemInstat : ProblemInstat
{
  DDProblemInstat(std::string name, ProblemStat& prob) 
    : ProblemInstat(name, prob),
      radius(1.0),
      epsilon(0.1),
      maxLevel(15),
      temperRefment(NULL),
      temperRefFct(NULL)
  { 
    Parameters::get("radius", radius);
    Parameters::get("epsilon", epsilon);
    center.set(0.0);
    Parameters::get("center", center);
    Parameters::get("mesh->max refinement level", maxLevel);
  }

  ~DDProblemInstat()
  {
    if (phase)
      delete phase;
    phase = NULL;
    
    if (phaseRefFct)
      delete phaseRefFct;
    phaseRefFct = NULL;
    
    if (phaseRefment)
      delete phaseRefment;
    phaseRefment = NULL;
    
    if (temperRefFct)
      delete temperRefFct;
    temperRefFct = NULL;
  
    if (temperRefment)
      delete temperRefment;
    temperRefment = NULL;
    
    if (c)
      delete c;
    c = NULL;
    
    if (ic)
      delete ic;
    ic = NULL;
  }

  void initialize(Flag initFlag,
		  ProblemInstat *adoptProblem = NULL,
		  Flag adoptFlag = INIT_NOTHING) override
  {
    problemStat->initialize(initFlag);

    ProblemInstat::initialize(INIT_NOTHING);
    
    c = new Circle(radius, center);
    ic = new InverseCircle(radius, center);
    
    phase = new DOFVector<double>(problemStat->getFeSpace(1), "phase2");
    phaseRefFct = new SignedDistRefinement(problemStat->getMesh(1));
    phaseRefment = new RefinementLevelCoords2(problemStat->getFeSpace(1), 
					      phaseRefFct, c);
    
    if (problemStat->getMeshes().size() == 2) {
      temperRefFct = new MySignedDistRefinement(problemStat->getMesh(0));
      temperRefment = new RefinementLevelCoords2(problemStat->getFeSpace(0), 
						 temperRefFct, c);
    }
  }

  

  void solveInitialProblem(AdaptInfo *adaptInfo) override
  {
    FUNCNAME("DDProblemInstat::solveInitialProblem");
    Timer t;
    
    SignedDistFctToPhaseField* p = 
      new SignedDistFctToPhaseField(epsilon, ic, 3.0);
    
    phaseRefment->refine(maxLevel);
    phase->interpol(p); 
//     phase->interpol(new One());
    *phase << valueOf(phase) + 1.e-5;

    
    if (problemStat->getMeshes().size() == 2) {
       temperRefment->refine(maxLevel, false, true); // when not use marker and estimator
      
//       int refLevel = 0;
//       Parameters::get("mesh->refinement->global heat refine", refLevel);
//       MSG("refLevel is %d\n", refLevel);
//       problemStat->getRefinementManager()->globalRefine(problemStat->getMesh(0), refLevel);
    }
    
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  MPI::COMM_WORLD.Barrier();
#endif
    MSG("mesh refinement and interpolation need %.5f seconds\n", t.elapsed());
    
    delete p;
  }


  DOFVector<double>* getPhase() { return phase; }

  double getEpsilon() { return epsilon; }


protected:

  double radius;
  double epsilon;
  int maxLevel;

  WorldVector<double> center;

  SignedDistRefinement* phaseRefFct;
  SignedDistRefinement* temperRefFct;

  RefinementLevelCoords2* phaseRefment;
  RefinementLevelCoords2* temperRefment;
  
  Circle* c;
  InverseCircle* ic;

  DOFVector<double>* phase;

};



// ===========================================================================

// ===== main program ========================================================

// ===========================================================================


int main(int argc, char* argv[])
{
  FUNCNAME("main");
  
  AMDiS::init(argc, argv);
  
  // ===== create and init the scalar problem ===== 

  ProblemStat heatSpace("heat");

  DDProblemInstat instat("heat", heatSpace);

  instat.initialize(INIT_ALL);

  // === create adapt info ===

  AdaptInfo adaptInfo("adapt", heatSpace.getNumComponents());

  // create instationary adapt

  AdaptInstationary adaptInstat("adapt",heatSpace,adaptInfo,instat,adaptInfo);
  
  
  // scale the mesh to a given size

  bool scaleMesh = false;

  Initfile::get("mesh->scale mesh",scaleMesh);

  if (scaleMesh) {
    WorldVector<double> scale; scale.set(1.0);
    Initfile::get("mesh->dimension",scale);
    for (int i = 0; i < heatSpace.getMeshes().size(); i++)
      Helpers::scaleMesh(heatSpace.getMesh(i), scale);
  }

  bool operatorOpt = true;
  Initfile::get("optimized operator", operatorOpt);


  // ===== create rhs functions =====

  
  int degree = heatSpace.getFeSpace()->getBasisFcts()->getDegree();
  F *rhsFct = new F(degree);
  rhsFct->setTimePtr(adaptInfo.getTimePtr());


  // ===== create operators =====
  

  // create laplace
  Operator A(heatSpace.getFeSpace());
  addSOT(A, valueOf(instat.getPhase()));
  heatSpace.addMatrixOperator(A, 0, 0);

  
  // create zero order operator
  Operator C(heatSpace.getFeSpace());
  addZOT(C, valueOf(instat.getPhase()));
  heatSpace.addMatrixOperator(C, 0, 0, instat.getInvTau(), instat.getInvTau());
  

#if USE_UH_OLD

  Operator C2(heatSpace.getFeSpace());
  C2.setUhOld(heatSpace.getSolution(0));
  addZOT(C2, valueOf(instat.getPhase()));
  heatSpace.addVectorOperator(C2, 0, instat.getInvTau(), instat.getInvTau());

#else

  Operator C2(heatSpace.getFeSpace());
  addZOT(C2, valueOf(heatSpace.getSolution(0)) * valueOf(instat.getPhase()));
  heatSpace.addVectorOperator(C2, 0, instat.getInvTau(), instat.getInvTau());
  
#endif

  Operator DBC(heatSpace.getFeSpace());
  addZOT(DBC, (36.0 / instat.getEpsilon()) * (1.0 - valueOf(instat.getPhase())));
  heatSpace.addMatrixOperator(DBC, 0, 0);

  
  // create RHS operator
  Operator F(heatSpace.getFeSpace());
  addZOT(F, valueOf(instat.getPhase()) * eval(rhsFct));
  heatSpace.addVectorOperator(F, 0);

  if (A.getNeedDualTraverse())
    MSG("A need dual traversal.\n");  
  if (C.getNeedDualTraverse())
    MSG("C need dual traversal.\n");
  if (C2.getNeedDualTraverse())
    MSG("C2 need dual traversal.\n");
  if (DBC.getNeedDualTraverse())
    MSG("DBC need dual traversal.\n");
  if (F.getNeedDualTraverse())
    MSG("F need dual traversal.\n");
  
  
  if (!operatorOpt) {
    A.useOptimizedAssembler(false);
    C.useOptimizedAssembler(false);
    C2.useOptimizedAssembler(false);
    DBC.useOptimizedAssembler(false);
    F.useOptimizedAssembler(false);
  }
    
    
    
  // ===== create boundary functions =====

  G *boundaryFct = new G;
//   boundaryFct->setTimePtr(instat.getTime());

  heatSpace.addDirichletBC(1, 0, 0, boundaryFct);

  heatSpace.addDirichletBC(3, 0, 0, boundaryFct);


#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  MPI::COMM_WORLD.Barrier();
#endif

  Timer t;
  // ===== start adaption loop =====

  adaptInstat.adapt();

  printf("whole process needs %.5f seconds\n", t.elapsed()); 
  
  delete rhsFct;rhsFct=NULL;
  delete boundaryFct;boundaryFct=NULL;
  
  AMDiS::finalize();
}



