/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#include "AMDiS.h"
#include "preconditioner/CahnHilliard_.h"
#include "Refinement.h"
#include "MeshFunction_Level.h"

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
#include "parallel/PetscSolverCahnHilliard.h"
#endif

using namespace AMDiS;

class MyCahnHilliard : public CahnHilliard_
{
public:
  MyCahnHilliard(std::string name)
  : CahnHilliard_(name), 
    refFunction(NULL),
    refinement(NULL) {}
  
  ~MyCahnHilliard()
  {
    if (refFunction) { delete refFunction; refFunction = NULL; }
    if (refinement) { delete refinement; refinement = NULL; }
  }
  
  void initData()
  {
    CahnHilliard_::initData();
    
    refFunction = new PhaseFieldRefinement(prob->getMesh());
    refinement = new RefinementLevelDOF(prob->getFeSpace(1), 
					refFunction,
					prob->getSolution()->getDOFVector(1));
    
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    Parallel::PetscSolverCahnHilliard* solver = dynamic_cast<Parallel::PetscSolverCahnHilliard*>(prob->getSolver());
    if (solver)
      solver->setChData2(&eps, &gamma, getTau());
#endif
  }
  
  /// Set initial condition and perform initial refinement
  void solveInitialProblem(AdaptInfo *adaptInfo)
  {
    // initial refinement
    refinement->refine(0);
    
    // refine until interfaces is resolved
    for (int i = 0; i < 5; ++i) {
      CahnHilliard_::solveInitialProblem(adaptInfo);
      refinement->refine(5);
    }

    CahnHilliard_::solveInitialProblem(adaptInfo);
  }

  /// Called at the end of each timestep.
  void closeTimestep(AdaptInfo *adaptInfo)
  {
    CahnHilliard_::closeTimestep(adaptInfo);
    refinement->refine(1);
  }
  
protected:
  PhaseFieldRefinement* refFunction;
  RefinementLevelDOF *refinement;
};



int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);  
  Timer t;
  
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  CreatorMap< LinearSolver >::addCreator("p_petsc_petsc-ch", new Parallel::PetscSolverCahnHilliard::Creator);
#endif
  
  MyCahnHilliard chProb("ch");
  chProb.initialize(INIT_ALL);
  chProb.initTimeInterface();

  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", chProb.getNumComponents());
  AdaptInstationary adaptInstat("adapt", chProb, adaptInfo, chProb, adaptInfo);

  int error_code = adaptInstat.adapt(); 

  MSG("elapsed time= %d sec\n", t.elapsed());
  AMDiS::finalize();

  return error_code;
}
