/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "AMDiS.h"
#include "NavierStokesPhase_TaylorHood.h"
#include "navierStokes.h"
#include "Refinement.h"
#include "MeshFunction_Level.h"
#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
#include "preconditioner/PetscSolverNavierStokes2.h"
#endif

using namespace AMDiS;

class NS_Channel : public NavierStokesPhase_TaylorHood
{
public:
  typedef NavierStokesPhase_TaylorHood super;
  
public:
  NS_Channel(std::string name_) : super(name_), zero_value(0.0), phaseField(NULL), inflowBC(NULL) {}
  
  ~NS_Channel()
  {
    if (phaseField != NULL) {
      delete phaseField;
      phaseField = NULL;
    }
    if (inflowBC != NULL) {
      delete inflowBC;
      inflowBC = NULL;
    }
  }
  
  void initData()
  {  
    super::initData();
  
    phaseField = new DOFVector<double>(getFeSpace(0), "phaseField"); 
    phaseField->set(1.0);
    super::setPhase(phaseField);
    
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    Parallel::PetscSolverNavierStokes2* solver = dynamic_cast<Parallel::PetscSolverNavierStokes2*>(prob->getSolver());
    if (solver) {
      solver->setStokesData(&viscosity, getInvTau(), prob->getSolution());
      solver->setPhase(phaseField);
    }
#endif
      
    double H = 4.1;
    double Um = 1.5;
    Parameters::get("mesh->H",H);
    Parameters::get("ns->Um",Um);
    inflowBC = new InflowBC(H,Um);
  }

  void solveInitialProblem(AdaptInfo *adaptInfo)
  {  
    super::solveInitialProblem(adaptInfo);

    double radius = 0.1;
    WorldVector<double> center;
    Parameters::get("obstacle->center",center);
    Parameters::get("obstacle->radius",radius);
    obstacle = new Circle(radius, center);
    
//     PhaseFieldRefinement refFunction(getMesh());
//     RefinementLevelDOF refinement(getFeSpace(), &refFunction, phaseField);
    SignedDistRefinement refFunction(getMesh());
    RefinementLevelCoords2 refinement(getFeSpace(), &refFunction, obstacle);
 
    // initial refinement
    refinement.refine(10);    
//     for (int i = 0; i < 5; i++) {
//       phaseField->interpol(new SignedDistFctToPhaseField(getEpsilon(), obstacle, -3.0));
//       refinement.refine(5);    
//     }
    phaseField->interpol(new SignedDistFctToPhaseField(getEpsilon(), obstacle, -3.0));
    
    if (inflowBC)
      inflowBC->setTimePtr(adaptInfo->getTimePtr());
  }
  
  void fillBoundaryConditions()
  {    
    WorldVector<double> zeroVec; zeroVec.set(0.0);
    AbstractFunction<WorldVector<double>, WorldVector<double> > *zeroVecFct = new AMDiS::Const<WorldVector<double>, WorldVector<double> >(zeroVec);

    super::setBcFct(zeroVecFct);    
    super::fillBoundaryConditions();
    
    
    // +------ 4 ------+
    // |               |
    // 1   o           2
    // |               |
    // +------ 3 ------+
    
    AbstractFunction<double, WorldVector<double> > *zero = new AMDiS::Const<double, WorldVector<double> >(0.0);
    
    /// at rigid wall: no-slip boundary condition
    for (size_t b = 3; b <= 6; b++) {
    for (size_t i = 0; i < dow; i++) {
      getProblem(0)->addDirichletBC(b, i, i, dynamic_cast<AbstractFunction<double, WorldVector<double> >*>(inflowBC));
    }
    }
    /// dirichlet bc for pressure at one DOF
//     getProblem(0)->addSingularDirichletBC(0, dow, dow, *zero);


    /// at left/right wall: prescribed velocity    
    if (inflowBC) {
      for (size_t b = 1; b <= 1; b++) {
	for (size_t i = 0; i < dow; i++)
	  getProblem(0)->addDirichletBC(b, i, i, dynamic_cast<AbstractFunction<double, WorldVector<double> >*>(inflowBC));
      }
    }
  }
  
private:
  
  double zero_value;
  
  DOFVector<double>* phaseField;
  AbstractFunction<double, WorldVector<double> >* obstacle;
  
  InflowBC* inflowBC;
};


int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);
  Timer t;
  
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  CreatorMap< LinearSolver >::addCreator("p_petsc_petsc-navierstokes2", new Parallel::PetscSolverNavierStokes2::Creator);
#endif

  NS_Channel nsProb("ns");
  nsProb.initialize(INIT_ALL | INIT_EXACT_SOLUTION);

  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", nsProb.getNumComponents());

  // adaption loop - solve ch-prob and ns-prob
  AdaptInstationary adaptInstat("adapt", nsProb, adaptInfo, nsProb, adaptInfo);
  
  // scale Mesh
  WorldVector<double> scale; scale.set(4.1); scale[0] = 25.0;
  Helpers::scaleMesh(nsProb.getMesh(), scale);

  nsProb.initTimeInterface();
  int error_code = adaptInstat.adapt(); 

  MSG("elapsed time= %f sec\n", t.elapsed());

  AMDiS::finalize();

  return error_code;
};
