/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#ifndef MTL_PFC_PRECONDITIONER_H
#define MTL_PFC_PRECONDITIONER_H

#include "AMDiS.h"
#include "solver/BlockPreconditioner.h"

#include <boost/numeric/mtl/interface/umfpack_solve.hpp>

namespace AMDiS { namespace extensions { 

using namespace std; 
using namespace AMDiS::MTLTypes;

template<typename MatrixType>
struct MTLPreconPfcBase : BlockPreconditioner<MatrixType>
{
  typedef BlockPreconditioner<MatrixType>   super;
  typedef typename super::precon_base       precon_base;
  typedef MTLPreconPfcBase<MatrixType>      self;
 
  class Creator : public CreatorInterfaceName<precon_base>
  {
  public:
    virtual ~Creator() {}
    
    precon_base* create() { 
      return new self(this->name);
    }
  };  
  
  MTLPreconPfcBase(std::string name)
    : super(),
      name(name), 
      tau(NULL),
      M0(1.0),
      matM(NULL), matL(NULL), matL0(NULL),
      PId(NULL), PDiagM(NULL), PDiagMpL(NULL), PDiagMpL2(NULL),
      solverM(NULL), solverMpL(NULL), solverMpL2(NULL)
  { }
  
  virtual ~MTLPreconPfcBase()
  {
    exit();
  }
  
  template<typename Data>
  void setData(Data* data)
  {
    tau = data->tau;
    M0 = data->M0;
    matM = &data->getM();
    matL = &data->getL();
    matL0 = &data->getL0();
  }
    
  void setData(double* tau_, double M0_ = 1.0)
  {
    tau = tau_;
    M0 = M0_;
  }

  
  virtual void init(const SolverMatrix<Matrix<DOFMatrix*> >& A, const MatrixType& fullMatrix) override;
  
  virtual void exit() override
  {    
    if (PId) { delete PId; PId = NULL; }
    if (PDiagM) { delete PDiagM; PDiagM = NULL; }
    if (PDiagMpL) { delete PDiagMpL; PDiagMpL = NULL; }
    if (PDiagMpL2) { delete PDiagMpL2; PDiagMpL2 = NULL; }
    if (solverM) { delete solverM; solverM = NULL; }
    if (solverMpL) { delete solverMpL; solverMpL = NULL; }
    if (solverMpL2) { delete solverMpL2; solverMpL2 = NULL; }
  }
  
  virtual void solve(const MTLVector& b, MTLVector& x) const override;  
    
  template <typename VectorX, typename VectorB>
  void solveM(VectorX& x, const VectorB& b, int nIter = -1) const
  {
    int nIterM=5;
    double rtolM=1.e-6,tolM=0;
    bool useDirectSolver = false;
    Parameters::get("precon_pfc_M->max iteration", nIterM);
    Parameters::get("precon_pfc_M->tolerance", tolM);
    Parameters::get("precon_pfc_M->relative tolerance", rtolM);
    Parameters::get("precon_pfc_M->use direct solver", useDirectSolver);
    if (useDirectSolver && solverM)
      (*solverM)(x,b);
    else {
      itl::basic_iteration<double> iter(b, nIterM, rtolM, tolM);
      x = 0.0;
      itl::cg(getM(), x, b, *PDiagM, *PId, iter);
    }
  }
  
  template <typename VectorX, typename VectorB>
  void solveMpL(VectorX& x, const VectorB& b) const
  {
    int nIterMpL=20;
    double rtolMpL=1.e-6,tolMpL=0;
    bool useDirectSolver = false;
    Parameters::get("precon_pfc_MpL->max iteration", nIterMpL);
    Parameters::get("precon_pfc_MpL->tolerance", tolMpL);
    Parameters::get("precon_pfc_MpL->relative tolerance", rtolMpL);
    Parameters::get("precon_pfc_MpL->use direct solver", useDirectSolver);
    if (useDirectSolver && solverMpL)
      (*solverMpL)(x,b);
    else {
      itl::basic_iteration<double> iter(b, nIterMpL, rtolMpL, tolMpL);
      x = 0.0;
      itl::cg(MpL, x, b, *PDiagMpL, *PId, iter);
    }
  }
  
  template <typename VectorX, typename VectorB>
  void solveMpL2(VectorX& x, const VectorB& b) const
  {
    int nIterMpL2=10;
    double rtolMpL2=1.e-6,tolMpL2=0;
    bool useDirectSolver = false;
    Parameters::get("precon_pfc_MpL2->max iteration", nIterMpL2);
    Parameters::get("precon_pfc_MpL2->tolerance", tolMpL2);
    Parameters::get("precon_pfc_MpL2->relative tolerance", rtolMpL2);
    Parameters::get("precon_pfc_MpL2->use direct solver", useDirectSolver);
    if (useDirectSolver && solverMpL2)
      (*solverMpL2)(x,b);
    else {
      itl::basic_iteration<double> iter(b, nIterMpL2, rtolMpL2, tolMpL2);
      x = 0.0;
      itl::cg(MpL2, x, b, *PDiagMpL2, *PId, iter);
    }
  }
  
  const MTLMatrix& getM() const { return matM ? *matM : super::A->getSubMatrix(2,2); }
  const MTLMatrix& getL0() const { return matL0 ? *matL0 : super::A->getSubMatrix(1,0); }
  const MTLMatrix& getL() const { return matL ? *matL : super::A->getSubMatrix(2,1); }
  
  double getTau() const { return *tau; }
  
protected: 
  std::string name;
  
  double* tau;
  double M0;
  
  MTLMatrix* matM;
  MTLMatrix* matL;
  MTLMatrix* matL0;
  
  MTLMatrix MpL;
  MTLMatrix MpL2;
   
  mutable MTLVector y0;
  mutable MTLVector y1;
  mutable MTLVector tmp;
  
  itl::pc::identity<MTLMatrix, double> *PId;
  itl::pc::diagonal<MTLMatrix, double> *PDiagM;
  itl::pc::diagonal<MTLMatrix, double> *PDiagMpL;
  itl::pc::diagonal<MTLMatrix, double> *PDiagMpL2;

  mtl::matrix::umfpack::solver<MTLMatrix> *solverM;
  mtl::matrix::umfpack::solver<MTLMatrix> *solverMpL;
  mtl::matrix::umfpack::solver<MTLMatrix> *solverMpL2;
};

typedef MTLPreconPfcBase<MTLTypes::MTLMatrix> MTLPreconPfc;

} }

#include "MTLPreconPfc.hh"

#endif // MTL_PFC_PRECONDITIONER_H

